/**
 *    @file
 *
 *    @section LICENSE
 *
 *    Copyright 2009 Bayer & Szell OEG
 *
 *    This file is part of secrond.
 *
 *    secrond is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    secrond is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with secrond.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    @section DESCRIPTION
 *
 *    Contains functions to implement typical daemon features.
 */

 
#include "daemon.h"


Boolean mkdirRecursive(char *path, mode_t mode) {
    char dir[TEMP_MAX];
    int i = 0;
    
    while (path[0] != '\0' && i < TEMP_MAX) {
        if (path[0] == '/' && i > 0) {
            dir[i] = '\0';
            if (mkdir(dir, mode) == -1 && errno != EEXIST) {
                return FALSE;
            }
        }
        dir[i] = path[0];
        i++;
        path++;
    }
    return TRUE;
}


void writeLog(Boolean doExit, Boolean noHeader, char *progName, int argSize, char *format, ...) {
    va_list args;
    time_t timestamp;
    struct tm *timeinfo;
    FILE *logFile;
    char *logFilePath, *msg, *buffer;
    int bufferSize;
    
    time(&timestamp);
    timeinfo = localtime(&timestamp);
    if (progName == NULL) {
        /* no prog name means output goes into the debug log */
        logFilePath = LOGFILE_DEBUG;
    } else {
        logFilePath = LOGFILE_EXEC;
    }
    logFile = fopen(logFilePath, "a");
    /** @todo alternate form of logging if opening/writing to/closing logfile fails? */
    if (logFile == NULL) {
        /* try to create the directory tree */
        if (mkdirRecursive(logFilePath, S_IRWXU)) {
            logFile = fopen(logFilePath, "a");
        }
        if (logFile == NULL) {
            return;
        }
    }
    /* generate formatted message */
    msg = (char *) malloc(my_strlen(format) + argSize + 1);
    if (msg != NULL) {
        va_start(args, format);
        vsprintf(msg, format, args);
        /* glue formatted message and header together (if any) */
        bufferSize = my_strlen(msg) + TEMP_MAX;
        if (progName != NULL) {
            bufferSize += my_strlen(progName);
        }
        buffer = (char *) malloc(bufferSize);
        if (buffer != NULL) {
            if (noHeader) {
                sprintf(buffer, "%s\n", msg);
            } else {
                if (progName == NULL) {
                    sprintf(buffer, "\n%s%s\n", asctime(timeinfo), msg);
                } else {
                    sprintf(buffer, "\n%s%s\n%s\n", asctime(timeinfo), progName, msg);
                }
            }
            /* output to log file */
            fwrite(buffer, my_strlen(buffer), 1, logFile);
            free(buffer);
        }
        free(msg);
        va_end(args);
    }
    fclose(logFile);
    if (doExit) {
        exit(EXIT_FAILURE);
    }
}


void daemonize() {
    pid_t pid;
    
    if (getuid() != ROOT_UID) {
        fprintf(stderr, "secrond: must be run as root\n");
        exit(EXIT_FAILURE);
    }
    signal(SIGHUP, SIG_IGN);
    pid = fork();
    if (pid < 0) {
        fprintf(stderr, "secrond: error forking daemon process\n");
        exit(EXIT_FAILURE);
    }
    if (pid > 0) {
        /* parent exits */
        exit(EXIT_SUCCESS);
    }
    if (setsid() < 0) {
        fprintf(stderr, "secrond: error detaching session\n");
        exit(EXIT_FAILURE);
    }
    if (chdir("/") < 0) {
        fprintf(stderr, "secrond: error changing working directory\n");
        exit(EXIT_FAILURE);
    }
    /* ignore stdin/out/err */
    if (freopen("/dev/null", "r", stdin) == NULL) {
        /* nothing */
    }
    if (freopen("/dev/null", "w", stdout) == NULL) {
        /* nothing */
    }
    if (freopen("/dev/null", "w", stderr) == NULL) {
        /* nothing */
    }
}


void singleInstance() {
    char pidBuffer[PID_MAX];
    int oldPid;
    
    lockFile = fopen(PIDFILE, "r+");
    if (lockFile == NULL) {
        /* only if the file does not exist, PIDFILE is openend in mode 'w' as to not erase its content */
        lockFile = fopen(PIDFILE, "w");
        if (lockFile == NULL) {
            writeLog(QUIT, NOAPPEND, NULL, my_strlen(PIDFILE), "cannot open %s - exiting ...", PIDFILE);
        }
    }
    if (lockf(fileno(lockFile), F_TLOCK, 0) < 0) {
        /* get process ID of other secrond instance if the PIDFILE is locked */
        if (fgets(pidBuffer, PID_MAX, lockFile) == NULL) {
            fclose(lockFile);
            lockFile = NULL;
            writeLog(QUIT, NOAPPEND, NULL, my_strlen(PIDFILE), "cannot read %s - exiting ...", PIDFILE);
        }
        oldPid = atoi(pidBuffer);
        fclose(lockFile);
        /* set lockFile to NULL so it is not deleted in freeResources() */
        lockFile = NULL;
        if (oldPid != 0) {
            writeLog(QUIT, NOAPPEND, NULL, PID_MAX, "another instance of secrond with process id %d seems to be running - exiting ...", oldPid);
        } else {
            writeLog(QUIT, NOAPPEND, NULL, my_strlen(PIDFILE), "cannot get file lock for %s - exiting ...", PIDFILE);
        }
    }
    sprintf(pidBuffer, "%d", (int) getpid());
    fwrite(pidBuffer, my_strlen(pidBuffer), 1, lockFile);
    fflush(lockFile);
    /* file is locked and has to be left open */
}


void handleSignalQuit(int s) {
    signal(s, handleSignalQuit);
    writeLog(NOQUIT, NOAPPEND, NULL, TEMP_MAX, "signal %d caught - exiting ...", s);
    if (s == SIGTERM) {
        exit(EXIT_SUCCESS);
    }
    exit(EXIT_FAILURE);
}


void handleSignalUser(int s) {
    signal(s, handleSignalUser);
    writeLog(NOQUIT, NOAPPEND, NULL, TEMP_MAX, "signal %d caught - reloading schedule ...", s);
    reloadSchedule = TRUE;
}


void handleSignals() {
    /* reload schedule on SIGUSR1 signal */
    signal(SIGUSR1, handleSignalUser);
    /* ignore any SIGHUPs or SIGCHLDs sent by parents or children processes */
    signal(SIGHUP, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    /* exit with log entry on error or termination signals */
    signal(SIGABRT, handleSignalQuit);
    signal(SIGFPE, handleSignalQuit);
    signal(SIGILL, handleSignalQuit);
    signal(SIGINT, handleSignalQuit);
    signal(SIGQUIT, handleSignalQuit);
    signal(SIGSEGV, handleSignalQuit);
    signal(SIGTERM, handleSignalQuit);
}


void handleSignalQuitChild(int s) {
    signal(s, handleSignalQuitChild);
    /* _exit does not call freeResources() */
    _exit(EXIT_FAILURE);
}


void handleSignalsChild() {
    signal(SIGUSR1, SIG_DFL);
    signal(SIGHUP, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    signal(SIGABRT, handleSignalQuitChild);
    signal(SIGFPE, handleSignalQuitChild);
    signal(SIGILL, handleSignalQuitChild);
    signal(SIGINT, handleSignalQuitChild);
    signal(SIGQUIT, handleSignalQuitChild);
    signal(SIGSEGV, handleSignalQuitChild);
    signal(SIGTERM, handleSignalQuitChild);
}
