/**
 *    @file
 *
 *    @section LICENSE
 *
 *    Copyright 2009 Bayer & Szell OEG
 *
 *    This file is part of secrond.
 *
 *    secrond is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    secrond is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with secrond.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    @section DESCRIPTION
 *
 *    Handles the execution of commands.
 */

 
#include "execute.h"


void execute(char *cmd) {
    pid_t child, grandchild;
    int pipes[2][2];
    FILE *progOutput;
    char *command = NULL, *outputBuffer = NULL;
    int c, chars = 0, bufferSize = 0, status;
    
    child = fork();
    if (child < 0) {
        writeLog(NOQUIT, NOAPPEND, NULL, my_strlen(cmd), "unable to fork child process for command %s", cmd);
        return;
    }
    if (child > 0) {
        /* parent returns to work */
        return;
    }
    /* forked child process */
    if (setsid() < 0) {
        writeLog(NOQUIT, NOAPPEND, NULL, my_strlen(cmd), "unable to detach session for command %s", cmd);
        _exit(EXIT_FAILURE);
    }
    /* reset to signal handlers using _exit instead of exit */
    handleSignalsChild();
    /* close file handlers we do not need */
    fclose(lockFile);
    /* copy cmd and free the schedule in the address space of the child */
    command = (char *) malloc(my_strlen(cmd) + 1);
    if (command == NULL) {
        writeLog(NOQUIT, NOAPPEND, NULL, TEMP_MAX, "error allocating %d bytes in memory", my_strlen(cmd) + 1);
        _exit(EXIT_FAILURE);
    }
    sprintf(command, "%s", cmd);
    freeDb(TRUE);
    /* open pipes for communication with grandchild */ 
    if (pipe(pipes[STDIN])) {
        writeLog(NOQUIT, NOAPPEND, NULL, my_strlen(command), "unable to create STDIN pipe for command %s", command);
        _exit(EXIT_FAILURE);
    }
    if (pipe(pipes[STDOUT])) {
        writeLog(NOQUIT, NOAPPEND, NULL, my_strlen(command), "unable to create STDOUT pipe for command %s", command);
        _exit(EXIT_FAILURE);
    }
    /* child may spawn grandchild using vfork, blocks child until exec call */
    grandchild = vfork();
    if (grandchild < 0) {
        writeLog(NOQUIT, NOAPPEND, NULL, my_strlen(command), "unable to vfork grandchild process for command %s", command);
        _exit(EXIT_FAILURE);
    }
    if (grandchild == 0) {
        /* vforked grandchild process, may use same address space as child until exec */
        /* close the ends of our pipes only the child is using */
        close(pipes[STDIN][WRITE_END]);
        close(pipes[STDOUT][READ_END]);
        /* make our pipes STDIN and STDOUT (STDIN unused) */
        /* we do not want to duplicate and close the descriptor if it is already where we want it */
        if (pipes[STDIN][READ_END] != STDIN) {
            dup2(pipes[STDIN][READ_END], STDIN);
            close(pipes[STDIN][READ_END]);
        }
        if (pipes[STDOUT][WRITE_END] != STDOUT) {
            dup2(pipes[STDOUT][WRITE_END], STDOUT);
            close(pipes[STDOUT][WRITE_END]);
        }
        dup2(STDOUT, STDERR);
        /* execute the command using our defined shell with environment var PATH */
        /** @todo do we want to provide more environment variables? */
        execlp(SHELL, SHELL, "-c", command, (char *) 0);
        /* unless exec fails, this section is unreachable */
        _exit(EXIT_FAILURE);
        /* grandchild dies */
    }
    /* child continues */
    /* close the ends of our pipes only the grandchild is using */
    close(pipes[STDIN][READ_END]);
    close(pipes[STDOUT][WRITE_END]);
    /* also close the STDIN write end since we will not be using it */
    close(pipes[STDIN][WRITE_END]);
    /* get output stream from STDOUT (and STDERR) */
    progOutput = fdopen(pipes[STDOUT][READ_END], "r");
    while ((c = getc(progOutput)) != EOF) {
        if (chars == bufferSize) {
            bufferSize += BUFFER_INC;
            outputBuffer = (char *) realloc(outputBuffer, bufferSize);
            if (outputBuffer == NULL) {
                writeLog(NOQUIT, NOAPPEND, NULL, TEMP_MAX, "error allocating %d bytes in memory", bufferSize);
                _exit(EXIT_FAILURE);
            }
        }
        outputBuffer[chars] = c;
        chars++;
    }
    fclose(progOutput);
    if (chars > 0) {
        /* output will be logged */
        if (chars == bufferSize) {
            bufferSize++;
            outputBuffer = (char *) realloc(outputBuffer, bufferSize);
            if (outputBuffer == NULL) {
                writeLog(NOQUIT, NOAPPEND, NULL, TEMP_MAX, "error allocating %d bytes in memory", bufferSize);
                _exit(EXIT_FAILURE);
            }
        }
        outputBuffer[chars] = '\0';
        writeLog(NOQUIT, NOAPPEND, command, chars, "%s", outputBuffer);
        free(outputBuffer);
    }
    /* make sure grandchild has finished, ignore interrupting signals */
    while (wait(&status) == -1 && errno == EINTR) {
        /* nothing */
    }
    /** @todo do we care about the return value of the grandchild? */
    free(command);
    /* _exit instead of exit so no user cleanup functions are called */
    _exit(EXIT_SUCCESS);
}
