/**
 *    @file
 *
 *    @section LICENSE
 *
 *    Copyright 2009 Bayer & Szell OEG
 *
 *    This file is part of secrond.
 *
 *    secrond is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    secrond is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with secrond.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    @section DESCRIPTION
 *
 *    Includes functions handling the parsing of the schedule file
 *    and adding the entries to the schedule.
 */


#include "parse.h"


/**
 * Tries to convert a string to an integer, and check it falls into the supplied min and max.
 *
 * @param[in]  str  String to convert.
 * @param[in]  min  Minimum value.
 * @param[in]  max  Maximum value.
 * @param[out] val  Converted integer value.
 *
 * @retval TRUE  No problems were encountered.
 * @retval FALSE @a str could not be converted or the value was not between min and max.
 */
static inline Boolean checkIntString(char *str, int min, int max, int *val) {
    if (str[0] == '\0') {
        return FALSE;
    }
    *val = atoi(str);
    if ((*val == 0 && str[0] != '0') || *val < min || *val > max) {
        /* converting the value failed or the result is invalid */
        return FALSE;
    }
    return TRUE;
}


/**
 * Sets the bits in SchedEntry time fields, when the entry would be executed.
 *
 * @param[in]  str  Time column of an entry in the schedule file.
 * @param[in]  min  Minimum time value.
 * @param[in]  max  Maximum time value.
 * @param[out] bits Bitstring with bits set accordingly.
 *
 * @retval TRUE  No problems were encountered.
 * @retval FALSE @a str could not be parsed.
 */
static Boolean setBits(char *str, int min, int max, bitstr_t *bits) {
    int i, j, k, step, val, valend, pos, pos2;
    char oneEntry[ONETIME_MAX];
    
    /* get the list of values */
    i = j = 0;
    for (; ; i++) {
        if (str[i] == '\0' || str[i] == ',') {
            /* an entry in the list is supposedly complete */
            if (j == 0) {
                return FALSE;
            }
            oneEntry[j] = '\0';
            if (oneEntry[0] == '/') {
                if (j == 1 || !checkIntString(&(oneEntry[1]), min, max, &step) || step == 0) {
                    return FALSE;
                }
                /* set step values */
                for (k = 0; k <= max - min; k += step) {
                    bit_set(bits, k);
                }
            }
            else if (oneEntry[0] == '*') {
                if (j != 1) {
                    return FALSE;
                }
                /* set all values */
                bit_nset(bits, 0, max - min);
            }
            else if ((pos = charpos(&(oneEntry[0]), '-')) != -1) {
                if (pos == 0 || pos == j - 1) {
                    return FALSE;
                }
                oneEntry[pos] = '\0';
                /* start */
                if (!checkIntString(&(oneEntry[0]), min, max, &val)) {
                    return FALSE;
                }
                /* check to see if there is a step value in addition to the range */
                if ((pos2 = charpos(&(oneEntry[pos + 1]), '/')) != -1) {
                    if (pos2 == 0 || pos2 == j - 1) {
                        return FALSE;
                    }
                    oneEntry[pos + 1 + pos2] = '\0';
                    /* step */
                    if (!checkIntString(&(oneEntry[pos + 1 + pos2 + 1]), min, max, &step) || step == 0) {
                        return FALSE;
                    }
                } else {
                    step = 1;
                }
                /* end */
                if (!checkIntString(&(oneEntry[pos + 1]), min, max, &valend)) {
                    return FALSE;
                }
                if (val > valend) {
                    return FALSE;
                }
                /* set (step) values in value range */
                for (k = val; k <= valend; k += step) {
                    bit_set(bits, k);
                }
            }
            else {
                if (!checkIntString(&(oneEntry[0]), min, max, &val)) {
                    return FALSE;
                }
                /* set one value */
                bit_set(bits, val - min);
            }
            j = 0;
            if (str[i] == '\0') {
                break;
            }
            continue;
        }
        /* ignore '*' in front of '/' */
        if (j == 1 && oneEntry[0] == '*' && str[i] == '/') {
            j--;
        }
        if (j >= ONETIME_MAX) {
            return FALSE;
        }
        /* add character to temporary entry string */
        oneEntry[j] = str[i];
        j++;
    }
    return TRUE;
}


/**
 * Adds an entry to the schedule.
 *
 * @param sec     Second string from the schedule file.
 * @param min     Minute string from the schedule file.
 * @param hour    Hour string from the schedule file.
 * @param mday    Day of the month string from the schedule file.
 * @param execCmd Command to be executed string from the schedule file.
 *
 * @retval TRUE  The entry was added successfully.
 * @retval FALSE An error occurred when trying to parse and add the entry.
 */
static Boolean addEntry(char *sec, char *min, char *hour, char *mday, char *execCmd) {
    SchedEntry *entry;
    #if DEBUG >= 1
    int debugFirstValue;
    #endif
    
    entry = (SchedEntry *) calloc(sizeof(SchedEntry), 1);
    if (entry == NULL) {
        writeLog(QUIT, NOAPPEND, NULL, TEMP_MAX, "error allocating %d bytes in memory", sizeof(SchedEntry) * sizeof(char));
    }
    entry->next = NULL;
    /* try to set all values of the entry */
    if (!setBits(sec, SEC_MIN, SEC_MAX, entry->sec)) {
        free(entry);
        return FALSE;
    }
    if (!setBits(min, MIN_MIN, MIN_MAX, entry->min)) {
        free(entry);
        return FALSE;
    }
    if (!setBits(hour, HOUR_MIN, HOUR_MAX, entry->hour)) {
        free(entry);
        return FALSE;
    }
    if (!setBits(mday, MDAY_MIN, MDAY_MAX, entry->mday)) {
        free(entry);
        return FALSE;
    }
    if (my_strlen(execCmd) == 0) {
        free(entry);
        return FALSE;
    }
    entry->execCmd = (char *) malloc(my_strlen(execCmd) + 1);
    if (entry->execCmd == NULL) {
        free(entry);
        writeLog(QUIT, NOAPPEND, NULL, TEMP_MAX, "error allocating %d bytes in memory", my_strlen(execCmd) + 1);
    }
    sprintf(entry->execCmd, "%s", execCmd);
    #if DEBUG >= 1
    writeLog(NOQUIT, NOAPPEND, NULL, my_strlen(entry->execCmd), "%s", entry->execCmd);
    bit_ffs(entry->sec, SEC_MAX - SEC_MIN + 1, &debugFirstValue);
    writeLog(NOQUIT, APPEND, NULL, TEMP_MAX, "first set second: %d", debugFirstValue + SEC_MIN);
    bit_ffs(entry->min, MIN_MAX - MIN_MIN + 1, &debugFirstValue);
    writeLog(NOQUIT, APPEND, NULL, TEMP_MAX, "first set minute: %d", debugFirstValue + MIN_MIN);
    bit_ffs(entry->hour, HOUR_MAX - HOUR_MIN + 1, &debugFirstValue);
    writeLog(NOQUIT, APPEND, NULL, TEMP_MAX, "first set hour: %d", debugFirstValue + HOUR_MIN);
    bit_ffs(entry->mday, MDAY_MAX - MDAY_MIN + 1, &debugFirstValue);
    writeLog(NOQUIT, APPEND, NULL, TEMP_MAX, "first set day: %d", debugFirstValue + MDAY_MIN);
    #endif
    /* add entry to the list */
    if (schedule->head == NULL) {
        schedule->head = schedule->tail = entry;
    } else {
        schedule->tail->next = entry;
        schedule->tail = entry;
    }
    schedule->entries++;
    return TRUE;
}


void setSchedule() {
    FILE *schedFile;
    char *pline;
    char line[LINE_MAX];
    char str_sec[TIMES_MAX], str_min[TIMES_MAX], str_hour[TIMES_MAX], str_mday[TIMES_MAX];
    int lineno = 0;
    
    freeDb(FALSE);
    schedFile = fopen(SCHEDFILE, "r");
    if (schedFile == NULL) {
        writeLog(QUIT, NOAPPEND, NULL, my_strlen(SCHEDFILE), "error opening %s", SCHEDFILE);
    }
    /* go through all lines of the SCHEDFILE and extract all values to call addEntry with */
    while (fgets(line, LINE_MAX, schedFile) != NULL) {
        lineno++;
        pline = &(line[0]);
        pline = skipWhitespace(pline);
        if (pline[0] == '#' || pline[0] == '\r' || pline[0] == '\n' || pline[0] == '\0') {
            continue;
        }
        pline = extractNonWhitespace(pline, TIMES_MAX, str_sec);
        pline = skipWhitespace(pline);
        pline = extractNonWhitespace(pline, TIMES_MAX, str_min);
        pline = skipWhitespace(pline);
        pline = extractNonWhitespace(pline, TIMES_MAX, str_hour);
        pline = skipWhitespace(pline);
        pline = extractNonWhitespace(pline, TIMES_MAX, str_mday);
        pline = skipWhitespace(pline);
        removeTrailingWhitespace(pline);
        if (!addEntry(str_sec, str_min, str_hour, str_mday, pline)) {
            writeLog(QUIT, NOAPPEND, NULL, my_strlen(SCHEDFILE) + TEMP_MAX, "error parsing %s at line %d", SCHEDFILE, lineno);
        }
    }
    fclose(schedFile);
}
