/**
 *    @file
 *
 *    @section LICENSE
 *
 *    Copyright 2009 Bayer & Szell OEG
 *
 *    This file is part of secrond.
 *
 *    secrond is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    secrond is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with secrond.  If not, see <http://www.gnu.org/licenses/>.
 *
 *    @section DESCRIPTION
 *
 *    Manages the schedule and knows when commands need to be executed.
 */

 
#include "secrond.h"


SchedEntryDb *schedule = NULL;
Boolean reloadSchedule = TRUE;
FILE *lockFile = NULL;


void freeDb(Boolean freeDbStructure) {
    SchedEntry *entry, *next;
    
    if (schedule == NULL) {
        return;
    }
    entry = schedule->head;
    while (entry != NULL) {
        next = entry->next;
        free(entry->execCmd);
        free(entry);
        entry = next;
    }
    schedule->head = schedule->tail = NULL;
    schedule->entries = 0;
    if (freeDbStructure) {
        free(schedule);
        schedule = NULL;
    }
}


void freeResources() {
    freeDb(TRUE);
    if (lockFile != NULL) {
        fclose(lockFile);
        remove(PIDFILE);
    }
}


/**
 * Determines whether an entry of the schedule is to be run, and calls execute if so.
 *
 * @param sec  Second part of the date/time to be checked.
 * @param min  Minute part of the date/time to be checked.
 * @param hour Hour part of the date/time to be checked.
 * @param mday Day of the month part of the date/time to be checked.
 */
static inline void execAt(int sec, int min, int hour, int mday) {
    SchedEntry *entry;
    
    entry = schedule->head;
    while (entry != NULL)  {
        #if DEBUG >= 3
        writeLog(NOQUIT, NOAPPEND, NULL, my_strlen(entry->execCmd), "checking %s ...", entry->execCmd);
        writeLog(NOQUIT, APPEND, NULL, TEMP_MAX, "match in second: %d", bit_test(entry->sec, sec - SEC_MIN));
        if (bit_test(entry->sec, sec - SEC_MIN)) {
            writeLog(NOQUIT, APPEND, NULL, TEMP_MAX, "match in minute: %d", bit_test(entry->min, min - MIN_MIN));
            if (bit_test(entry->min, min - MIN_MIN)) {
                writeLog(NOQUIT, APPEND, NULL, TEMP_MAX, "match in hour: %d", bit_test(entry->hour, hour - HOUR_MIN));
                if (bit_test(entry->hour, hour - HOUR_MIN)) {
                    writeLog(NOQUIT, APPEND, NULL, TEMP_MAX, "match in day: %d", bit_test(entry->mday, mday - MDAY_MIN));
                }
            }
        }
        #endif
        /* if all relevant values of the current time are set in the bitstrings of an entry, it is executed */
        if (bit_test(entry->sec, sec - SEC_MIN) &&
            bit_test(entry->min, min - MIN_MIN) &&
            bit_test(entry->hour, hour - HOUR_MIN) &&
            bit_test(entry->mday, mday - MDAY_MIN)) {
            #if DEBUG >= 2
            writeLog(NOQUIT, NOAPPEND, NULL, my_strlen(entry->execCmd), "executing %s ...", entry->execCmd);
            #endif
            #if SIMULATE <= 0
            execute(entry->execCmd);
            #endif
        }
        entry = entry->next;
    }
}


/**
 * Sets up the daemon and runs execAt periodically. 
 *
 * @param argc Amount of command-line arguments.
 * @param argv Command-line arguments, ignored.
 *
 * @retval 0 The program exited without errors due to a @e SIGTERM signal.
 * @retval 1 The program aborted due to an error.
 */
int main(int argc, char *argv[]) {
    time_t timestamp, previoust;
    struct tm *timeinfo;
    
    daemonize();
    atexit(freeResources);
    handleSignals();
    writeLog(NOQUIT, NOAPPEND, NULL, 0, "secrond starting ...");
    singleInstance();
    schedule = (SchedEntryDb *) malloc(sizeof(SchedEntryDb));
    if (schedule == NULL) {
        writeLog(QUIT, NOAPPEND, NULL, TEMP_MAX, "error allocating %d bytes in memory", sizeof(SchedEntryDb));
    }
    schedule->head = schedule->tail = NULL;
    schedule->entries = 0;
    /**
     * The main loop gets the current time and compares it with the one from the previous run.
     * If the current time is less than the previous one, it is slept until the current time is greater again.
     * If the current time is greater than the previous one, it goes through all seconds in between and finds out what to execute at that time.
     * For both statements we assume that we cannot lose or win more than TIMECHANGE_MAX seconds between two runs.
     * One second is slept between runs.
     * @todo An optimization might be to sleep longer than one second and really only wake up when there is something to execute.
     */
    time(&previoust);
    previoust--;
    while (TRUE) {
        if (reloadSchedule == TRUE) {
            setSchedule();
            reloadSchedule = FALSE;
        }
        time(&timestamp);
        if (timestamp - previoust > TIMECHANGE_MAX || /* too much time ahead */
            timestamp - previoust < -TIMECHANGE_MAX) { /* too much time behind */
            previoust = timestamp;
        }
        if (timestamp <= previoust) {
            sleep(1);
            continue;
        }
        for (previoust++; previoust <= timestamp; previoust++) {
            timeinfo = localtime(&previoust);
            execAt(timeinfo->tm_sec, timeinfo->tm_min, timeinfo->tm_hour, timeinfo->tm_mday);
        }
        previoust--;
        sleep(1);
    }
    /* UNREACHABLE */
}
